/*====================================================================================================*
* PINEWOOD DERBY - PWD 3-Lane Track Timer                                  Version 1.1 - 25 Feb. 2022 *
* Created by: Bill van Leeuwen (email: Billv923@outlook.com) for the Arduino UNO microcontroller      *
*                                                                                                     *
* This program is designed to work in conjunction with the Apple Macintosh based Pinewood Master race *
* management software by Marc Speth which runs on a MAC attached to the Arduino via the USB serial    *
* port. This is accomplished by emulating the Micro Wizard(tm) Fast Track K2 series command and       *
* communication data format. Note that in this version the GATE switch must be held open while the    *
* race is in progress. Closing the GATE pre-maturely will reset the timer.                            *
*                                                                                                     *
* This program measures the start-to-finish time of 3 lanes. Race times and finish order are sent to  *
* the serial port to interface with the race management/display software running on the MAC PC. Time  *
* is measured from activation of the start switch to activation of the finish line sensor of each     *
* lane. Each car has 10 seconds to finish the race. If a car doesn't cross the finish line within 10  *
* seconds, it is assumed that the car had a problem or the lane was not used. In either case, the     *
* value 9.999 is assigned to the lane time as an error indication for use by the race management      *
* software running on the MAC.                                                                        *
*                                                                                                     *
* This design uses optical sensors (one for each lane) embedded in the track at the finish line to    *
* detect when a car crosses the finish line. Incandescent lamps or other infrared light source        *
* mounted above the finish line sensors provide the illumination needed by the sensors. A switch      *
* connected to the start gate is used to start the timer.                                             *
*                                                                                                     *
* Refer to the Schematic Diagram, parts list and design notes that accompany this software for wiring *
* of the infrared lane sensors, race start switch, and reset switch.                                  *
* Feel free to contact me at Billv923@outlook.com if any questions.                                   *
*====================================================================================================*/

//Initialization
#define Num_Lanes    3         //Number of lanes
#define NotReady     0         //Race State = Not Ready
#define Ready        1         //Race State = Ready
#define Racing       2         //Race State = Racing
#define Finished     3         //Race State = Finished
#define Msg_Reset   'R'        //Reset character received from PC via Serial port
#define SMsg_Lmask  'M'        //Mask Lane command received from PC via Serial port

//Pin Assignments
#define fs1          2         //Finish Sensor Lane 1 - Digital Input Pin 2
#define fs2          3         //Finish Sensor Lane 2 - Digital Input Pin 3
#define fs3          4         //Finish Sensor Lane 3 - Digital Input Pin 4
#define gate_sw      8         //Gate Switch (Start)  - Digital Input Pin 8
#define reset_sw     9         //Manual Reset         - Digital Input Pin 9
#define Rled        10         //Timer NOT READY indicator
#define Yled        11         //Timer READY indicator
#define Gled        12         //Timer RACING indicator
#define Pin13       13         //IO Pin 13 LED        - Digital Pin 13


//Variable Initialization
int i, n;                      //Working variables
int lane_place[Num_Lanes];     //Lane finish place array
boolean track_ready;           //False if any lane sensor is obstructed or Gate Switch Open
boolean Lane1_Mask;            //True if lane 1 is masked
boolean Lane2_Mask;            //True if lane 2 is masked
boolean Lane3_Mask;            //True if lane 3 is masked
byte regData;                  //Pins D0 - D7 'D' register input data
int Active_Lanes;              //Number of active lanes (Total lanes - masked lanes)
int Masked_Lanes;              //Number of lanes that are masked
int lane;                      //Lane number
int finish_order;              //Finish order
int serial_data;               //Serial data
int racestate;                 //0=NotReady, 1=Ready, 2=Racing, 3=Finished
int fincount;                  //Tallies cars crossing finish line to determine race over
unsigned long Start_Time;      //Time (microseconds)
unsigned long Exp_Time;        //Expiration Time (milliseconds)
unsigned long Clk_Time;        //CPU Clock Time (microseconds)
float Time;                    //Working variable - Calculated Lane Time (seconds)
float lane_time[Num_Lanes];    //Calculated lane time data array (seconds)
String Lane_Array[6] = {"A=", "B=", "C=", "D=", "E=", "F="};  //Lane identifiers used in serial message
char Fin_Array[6] = {'!', '"', '#', '$', '%', '&'}; //Finish Order characters used in serial message

//======================================= SETUP ROUTINE =======================================
void setup() {  
//Hardware Setup
  //Initialize Finish Sensor, Gate Switch and Reset Switch pins as inputs & LED pins as outputs
  pinMode(fs1,INPUT_PULLUP);          //Note: Sensor inputs fs1 thru fs3 transition from LOW to
  pinMode(fs2,INPUT_PULLUP);          //      HIGH when light beam is interupted (e.g. car
  pinMode(fs3,INPUT_PULLUP);          //      crosses finish line)
  pinMode(gate_sw,INPUT_PULLUP);      //Gate_sw is normally LOW & transitions to HIGH at race start
  pinMode(reset_sw,INPUT_PULLUP);     //Reset_sw is normally HIGH & transitions to LOW when pressed
  pinMode(Rled, OUTPUT);              //Set Rled pin as an output
  pinMode(Yled, OUTPUT);              //Set Yled pin as an output
  pinMode(Gled, OUTPUT);              //Set Gled pin as an output
  pinMode(Pin13, OUTPUT);             //set as output so we can turn off the pin 13 LED
  
//Software Setup
  Serial.begin(9600);                //Enable serial comms to PC at 9600 baud
  racestate = NotReady;              //Set initial race state to not ready
  Lane1_Mask = false;                //Preset Lane 1 mask flag to false
  Lane2_Mask = false;                //Preset Lane 2 mask flag to false
  Lane3_Mask = false;                //Preset Lane 3 mask flag to false
  Masked_Lanes = 0;                  //Preset masked lane count to zero
  digitalWrite(Pin13,LOW);           //Turn off Pin 13 LED just for the heck of it
  Set_State_Indicator();             //Turn on red NOT READY state indicator
}

//======================================== MAIN LOOP ============================================
void loop() {
  //READ & PROCESS INCOMING SERIAL DATA
  serial_data = get_serial_data();                //Get any received serial data
  if (serial_data == int(Msg_Reset)) {            //Check for Reset request from PC
    Check_Track();                                //If so, call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Set_State_Indicator();                      //Turn on yellow READY state indicator
    } 
    else {                                        //Do this if track is not ready
      racestate = NotReady;                       //Set racestate to Not Ready
      Set_State_Indicator();                      //Turn on red NOT READY state indicator
    }
  }
  if (serial_data == int(SMsg_Lmask)) {           //Check for lane Mask request from PC
    delay(10);                                    //wait to ensure next serial character is received from GPRM
    serial_data = get_serial_data();              //Get serial data
    if (serial_data == int('A'))   {              //Check if Lane 1 is to be masked
      Lane1_Mask = true;                          //Set to TRUE if so
      ++Masked_Lanes;                             //Increment masked lane count
    }
    if (serial_data == int('B')) {                //Check if Lane 2 is to be masked
      Lane2_Mask = true;                          //Set to TRUE if so
      ++Masked_Lanes;                             //Increment masked lane count
    }
    if (serial_data == int('C')) {                //Check if Lane 3 is to be masked
      Lane3_Mask = true;                          //Set to TRUE if so
      ++Masked_Lanes;                             //Increment masked lane count
    }
    if (serial_data == int('G')) {                //Check if all lanes are to be unmasked
      Masked_Lanes = 0;                           //Clear masked lane count
      Lane1_Mask = false;                         //Clear Lane 1 mask flag
      Lane2_Mask = false;                         //Clear Lane 2 mask flag
      Lane3_Mask = false;                         //Clear Lane 3 mask flag
    }
    Serial.println("*");                          //Send mask acknowledge character to PC
  }
 
  //CHECK & PROCESS RACE STATES
  //Process activation of the Manual Reset switch
  if (digitalRead(reset_sw)==0){                  //Manual timer Reset button pressed?
    delay(500);                                   //Insert delay to debounce reset switch
    Lane1_Mask = false;                           //Clear Lane 1 mask flag in case it was set
    Lane2_Mask = false;                           //Clear Lane 2 mask flag in case it was set
    Lane3_Mask = false;                           //Clear Lane 3 mask flag in case it was set
    Masked_Lanes = 0;                             //Clear masked lane count in case it was set
    Check_Track();                                //Call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Set_State_Indicator();                      //Turn on yellow READY state indicator
    } 
    else {                                        //Do this if track is not ready
      racestate = NotReady;                       //Set racestate to NotReady
      Set_State_Indicator();                      //Turn on red NOT READY state indicator
    }
  }
  if (racestate == NotReady) {                    //Monitor track status loop
    Check_Track();                                //Call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      racestate = Ready;                          //Set racestate to Ready
      Set_State_Indicator();                      //Turn on yellow READY state indicator
    } 
    delay(1000);                                  //Wait 1 sec. before we check track again
  }
  // Transition from Ready to Racing when the Gate switch is tripped
  if (racestate == Ready && digitalRead(gate_sw)==1) {
    Start_Time = micros();                        //Get start time from CPU clock in microseconds
    Exp_Time = Start_Time + 10000000;             //Set Expiration = Start_Time + 10 seconds (in microseconds)
    racestate = Racing;                           //Set racestate to Racing
    Set_State_Indicator();                        //Turn on green RACING state indicator
    delay(200);                                   //Wait a while before we start finish check loop
    
    //Race underway. Check/Wait for everyone to finish
    /*====================================================================================================*
    * NOTE!!!!: The 'while' loop below has been optimized to repeatedly check lane sensors as fast as     *
    * possible by accessing the digital port input register directly via the PIND command instead of the  *
    * 'digitalRead()' command. This command is specifically tailored to Arduino boards using the ATmega8, *
    * ATmega168 or ATmega328p chips. If this code is used with an Arduino clone having a different        *
    * chip, it may not work.                                                                              *
    *====================================================================================================*/
    Active_Lanes = Num_Lanes - Masked_Lanes;      //Calculate number of active lanes
    while (fincount<Active_Lanes) {               //Loop until all cars have crossed finish line
      Clk_Time = micros();                        //Get the CPU clock time in microseconds
      regData = PIND;                             //Parallel read the D0-D7 data register
      if (lane_place[0]==0) {                     //Check & skip if Lane 1 already finished
        if (!Lane1_Mask) {                        //Check & skip if Lane 1 is masked
          if (regData & 0x04) {                   //Is Car over Lane 1 Finish Line?
            lane_time[0]=Clk_Time;                //Yes...get time
            lane_place[0]=++fincount;             //Incr finish count & set finish place for Lane 1
          }
        }
      }
      if (lane_place[1]==0) {                     //Check & skip if Lane 2 already finished
        if (!Lane2_Mask) {                        //Check & skip if Lane 2 is masked
          if (regData & 0x08) {                   //Is Car over Lane 2 Finish Line?
            lane_time[1]=Clk_Time;                //Yes...get time
            lane_place[1]=++fincount;             //Incr finish count & set finish place for Lane 2
          }
        }
      }  
      if (lane_place[2]==0) {                     //Check & skip if Lane 3 already finished
        if (!Lane3_Mask) {                        //Check & skip if Lane 3 is masked
          if (regData & 0x10) {                   //Is Car over Lane 3 Finish Line?
            lane_time[2]=Clk_Time;                //Yes...get time
            lane_place[2]=++fincount;             //Incr finish count & set finish place for Lane 3
          }
        }
      }
      if (Clk_Time>Exp_Time) break;               //Exit racing mode if time has expired
      if (digitalRead(gate_sw)==0) break;         //Exit racing mode if gate is closed to end the race
    }
    //Race is over or time has expired
    racestate=Finished;                           //Set racestate to Finished
    Set_State_Indicator();                        //Turn on red NOT READY state indicator
    //Send Race Results To PC as one long string in FAST TRACK(tm) format. Race results string
    //format is as follows: A=N.NNNX B=N.NNNX C=N.NNNX D=N.NNNX E=N.NNNX F=N.NNNX <cr><lf>
    //Where: A - F reference lanes 1 - 6 respectively
    //       N.NNN is the lane time in seconds
    //       X is the finish sequence character as follows:
    //         ! = 1'st place, " = 2'nd place, # = 3'd place, $ = 4'th place, % = 5'th place, & = 6'th place
    for (n=0; n<Num_Lanes; n++) {                 //Loop thru all lane times
      Serial.print(Lane_Array[n]);                //Print lane prefix
      if (lane_time[n]>0) {                       //Check to see if lane time is valid
        Time = (float)((lane_time[n]-Start_Time)/1000000.0);  //Calc lane time and convert to seconds
        Serial.print(Time,3);                     //Print lane time trimmed to 3 places after decimal
      }
      else {
        Serial.print("9.999");                    //Print all nines if lane time is invalid
      }
      i = lane_place[n];                          //Get the lane place (finish order)
      Serial.print(Fin_Array[i-1]);               //And print associated finish sequence character
      Serial.print(" ");                          //Print a space separator
    }
    Serial.println("D=0.000  E=0.000  F=0.000 "); //Send filler for 3 unused lanes (4, 5 & 6) + a cr/lf
    
    delay(200);                                   //Wait a while to allow display SW to process times
  }                                               //End Racing "if"

  // Transition from Finished to Ready when the Gate switch is reset (closed)
  if (racestate == Finished && digitalRead(gate_sw)==0) {
    delay(500);                                   //Wait 1/2 sec. to debounce gate switch
    Serial.println("@");                          //Send the '@' symbol to indicate start gate was closed
    initialize();                                 //Call initialize function & reset everything
    racestate = NotReady;                         //Set racestate to Not Ready
    Set_State_Indicator();                        //Turn on NOT READY state indicator
  }
}                                                 //Loop Back

/*================================================================================*
  SUBROUTINES & FUNCTIONS
 *================================================================================*/
//Subroutine: Initialize Racing Variables
void initialize() {  
  for (int n=0; n<Num_Lanes; n++) {           //Reset lane Place & Times
    lane_place[n] = 0;
    lane_time[n] = 0;
  }
  fincount = 0;                               //Reset finish count to 0
  Serial.flush();                             //Clear the serial input buffer
  return;
}

//Function: Read Serial Data
int get_serial_data() {
  int data = 0;                               //Clear any old data
  if (Serial.available() > 0) {               //Check for new data
    data = Serial.read();                     //and get it if so
  }
  return data;
}  

//Function: Check Track - Sets 'track_ready' to true or false
int Check_Track() {
  track_ready = true;                         //Default track ready status to true
  if (digitalRead(gate_sw)==1) {              //Check if Gate Switch is open
    track_ready = false;                      //And set track ready status to false
  }
  for (int n=0; n<Num_Lanes; n++) {           //Loop through all 3 lanes and
    if (digitalRead(n+2)==1) {                //Check if a lane sensor (pins 2-4) is obstructed
      track_ready = false;                    //If so, set track ready status to false
    }
  }
}

void Set_State_Indicator() {
  digitalWrite(Rled,HIGH);                    //Turn off NOT READY LED
  digitalWrite(Yled,HIGH);                    //Turn off READY LED
  digitalWrite(Gled,HIGH);                    //Turn off RACING LED
  switch(racestate) {
    case 0:
      digitalWrite(Rled,LOW);                //Turn on NOT READY LED
      break;
    case 1:
      digitalWrite(Yled,LOW);                //Turn on READY LED
      break;
    case 2:
      digitalWrite(Gled,LOW);                //Turn on RACING LED
      break;
    case 3:
      digitalWrite(Rled,LOW);                //Turn on NOT READY LED
      break;
    default:
      //Do nothing
      break;
  }
}
